#include <glad/glad.h>
#include <GL/freeglut.h>

#include <string>
#include <iostream>
#include <fstream>
#include <vector>

#define STB_IMAGE_IMPLEMENTATION
#include "stb_image.h"

#include <glm/gtc/matrix_transform.hpp>

typedef unsigned short  ushort;
typedef glm::vec2       vec2;
typedef glm::vec3       vec3;
typedef glm::mat4       mat4;
typedef glm::vec4       color4;

struct Vertex
{
    vec3 pos;
    vec2 texcoord;
    vec3 normal;
};

int g_width = 256;
int g_height = 256;
unsigned int g_texture;

void showErrorMessageAndExit(std::string message)
{
    std::cout << message << std::endl;
    std::cout << "Press the Enter key to exit." << std::endl;
    std::getchar();
    exit(1);
}

std::string readStringFromFile(std::string filename)
{
    std::string res;
    std::string line;
    std::ifstream in(filename);

    if (in.is_open())
    {
        while (getline(in, line))
        {
            res += line + "\n";
        }
    }
    in.close();

    return res;
}

GLint CompileShader(std::string filename)
{
    GLint vs = glCreateShader(GL_VERTEX_SHADER);
    std::string cString = readStringFromFile("Data\\Shaders\\" + filename + ".vert.glsl");
    GLchar const* files[] = { cString.c_str() };
    GLint lengths[] = { cString.size() };
    glShaderSource(vs, 1, files, lengths);
    glCompileShader(vs);

    GLint success;
    GLchar infoLog[512];
    glGetShaderiv(vs, GL_COMPILE_STATUS, &success);
    if (!success)
    {
        glGetShaderInfoLog(vs, 512, NULL, infoLog);
        showErrorMessageAndExit("Vertex Shader Error. " + std::string(infoLog));
    }

    GLint fs = glCreateShader(GL_FRAGMENT_SHADER);
    cString = readStringFromFile("Data\\Shaders\\" + filename + ".frag.glsl");
    GLchar const* files1[] = { cString.c_str() };
    GLint lengths1[] = { cString.size() };
    glShaderSource(fs, 1, files1, lengths1);
    glCompileShader(fs);

    glGetShaderiv(fs, GL_COMPILE_STATUS, &success);
    if (!success)
    {
        glGetShaderInfoLog(fs, 512, NULL, infoLog);
        showErrorMessageAndExit("Fragment Shader Error. " + std::string(infoLog));
    }

    GLint prog = glCreateProgram();
    glAttachShader(prog, vs);
    glAttachShader(prog, fs);
    glLinkProgram(prog);
    glUseProgram(prog);

    glDetachShader(prog, vs);
    glDetachShader(prog, fs);
    glDeleteShader(vs);
    glDeleteShader(fs);

    return prog;
}

void initVertexBuffers(GLint program)
{
    //    v6----- v5
    //   /|      /|
    //  v1------v0|
    //  | |     | |
    //  | |v7---|-|v4
    //  |/      |/
    //  v2------v3

    std::vector<Vertex> verts(12);

    // Vertices

    // v0-v1-v2-v3 front
    verts[0].pos = vec3(0.5f, 0.5f, 0.5f);
    verts[1].pos = vec3(-0.5f, 0.5f, 0.5f);
    verts[2].pos = vec3(-0.5f, -0.5f, 0.5f);
    verts[3].pos = vec3(0.5f, -0.5f, 0.5f);
    // v0-v3-v4-v5 left
    verts[4].pos = vec3(0.5f, 0.5f, 0.5f);
    verts[5].pos = vec3(0.5f, -0.5f, 0.5f);
    verts[6].pos = vec3(0.5f, -0.5f, -0.5f);
    verts[7].pos = vec3(0.5f, 0.5f, -0.5f);
    // v0-v5-v6-v1 top
    verts[8].pos = vec3(0.5f, 0.5f, 0.5f);
    verts[9].pos = vec3(0.5f, 0.5f, -0.5f);
    verts[10].pos = vec3(-0.5f, 0.5f, -0.5f);
    verts[11].pos = vec3(-0.5f, 0.5f, 0.5f);

    // Texture Coordinates

    // v0-v1-v2-v3 front
    verts[0].texcoord = vec2(1.0f, 1.0f);
    verts[1].texcoord = vec2(0.0f, 1.0f);
    verts[2].texcoord = vec2(0.0f, 0.0f);
    verts[3].texcoord = vec2(1.0f, 0.0f);
    // v0-v3-v4-v5 left
    verts[4].texcoord = vec2(0.0f, 1.0f);
    verts[5].texcoord = vec2(0.0f, 0.0f);
    verts[6].texcoord = vec2(1.0f, 0.0f);
    verts[7].texcoord = vec2(1.0f, 1.0f);
    // v0-v5-v6-v1 top
    verts[8].texcoord = vec2(1.0f, 0.0f);
    verts[9].texcoord = vec2(1.0f, 1.0f);
    verts[10].texcoord = vec2(0.0f, 1.0f);
    verts[11].texcoord = vec2(0.0f, 0.0f);

    // Normals
    // v0-v1-v2-v3 front
    verts[0].normal = vec3(0.0f, 0.0f, 1.0f);
    verts[1].normal = vec3(0.0f, 0.0f, 1.0f);
    verts[2].normal = vec3(0.0f, 0.0f, 1.0f);
    verts[3].normal = vec3(0.0f, 0.0f, 1.0f);
    // v0-v3-v4-v5 left
    verts[4].normal = vec3(1.0f, 0.0f, 0.0f);
    verts[5].normal = vec3(1.0f, 0.0f, 0.0f);
    verts[6].normal = vec3(1.0f, 0.0f, 0.0f);
    verts[7].normal = vec3(1.0f, 0.0f, 0.0f);
    // v0-v5-v6-v1 top
    verts[8].normal = vec3(0.0f, 1.0f, 0.0f);
    verts[9].normal = vec3(0.0f, 1.0f, 0.0f);
    verts[10].normal = vec3(0.0f, 1.0f, 0.0f);
    verts[11].normal = vec3(0.0f, 1.0f, 1.0f);

    std::vector<ushort> inds(18);
    // v0-v1-v2-v3 front
    inds[0] = 0;
    inds[1] = 1;
    inds[2] = 2;
    inds[3] = 0;
    inds[4] = 2;
    inds[5] = 3;
    // v0-v3-v4-v5 left
    inds[6] = 4;
    inds[7] = 5;
    inds[8] = 6;
    inds[9] = 4;
    inds[10] = 6;
    inds[11] = 7;
    // v0-v5-v6-v1 top
    inds[12] = 8;
    inds[13] = 9;
    inds[14] = 10;
    inds[15] = 8;
    inds[16] = 10;
    inds[17] = 11;

    mat4 projMatrix = glm::perspective(70.0f * 3.14f / 180.0f, (float)g_width / g_height, 0.1f, 100.0f);
    GLint uProjMatrixLocation = glGetUniformLocation(program, "uProjMatrix");
    glUniformMatrix4fv(uProjMatrixLocation, 1, GL_FALSE, &projMatrix[0][0]);

    mat4 viewMatrix = glm::lookAt(vec3(0.8f, 0.7f, 1.1f), vec3(0.0f), vec3(0, 1, 0));
    GLint uViewMatrixLocation = glGetUniformLocation(program, "uViewMatrix");
    glUniformMatrix4fv(uViewMatrixLocation, 1, GL_FALSE, &viewMatrix[0][0]);

    GLuint vertexBufferObject;
    GLuint indexBufferObject;
    glGenBuffers(1, &vertexBufferObject);
    glGenBuffers(1, &indexBufferObject);

    glBindBuffer(GL_ARRAY_BUFFER, vertexBufferObject);
    glBufferData(GL_ARRAY_BUFFER, sizeof(Vertex) * verts.size(), &(verts[0].pos.x), GL_STATIC_DRAW);

    GLint aPositionLoc = glGetAttribLocation(program, "aPosition");
    if (aPositionLoc == -1) showErrorMessageAndExit("Failed to get a \"aPosition\" variable.");
    glVertexAttribPointer(aPositionLoc, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (const GLvoid*)(offsetof(Vertex, pos)));
    glEnableVertexAttribArray(aPositionLoc);

    GLint aTexLoc = glGetAttribLocation(program, "aTexCoord");
    if (aTexLoc == -1) showErrorMessageAndExit("Failed to get a \"aTexCoord\" variable.");
    glVertexAttribPointer(aTexLoc, 2, GL_FLOAT, GL_FALSE, sizeof(Vertex), (const GLvoid*)(offsetof(Vertex, texcoord)));
    glEnableVertexAttribArray(aTexLoc);

    GLint aNormalLoc = glGetAttribLocation(program, "aNormal");
    if (aNormalLoc == -1) showErrorMessageAndExit("Failed to get a \"aNormal\" variable.");
    glVertexAttribPointer(aNormalLoc, 3, GL_FLOAT, GL_FALSE, sizeof(Vertex), (const GLvoid*)(offsetof(Vertex, normal)));
    glEnableVertexAttribArray(aNormalLoc);

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexBufferObject);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(ushort) * inds.size(), &(inds[0]), GL_STATIC_DRAW);
}

void initTextures()
{
    glEnable(GL_TEXTURE_2D);

    int width, height, nrChannels;
    const char* texturePath = "./Data/Textures/box.png";
    unsigned char *data = stbi_load(texturePath, &width, &height, &nrChannels, 0);
    if (!data)
    {
        showErrorMessageAndExit("Failed to load the texture: " + std::string(texturePath));
    }

    glGenTextures(1, &g_texture);
    glBindTexture(GL_TEXTURE_2D, g_texture);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

    stbi_image_free(data);
}

void initialize()
{
    // Load definitions for OpenGL functions
    if (!gladLoadGL())
    {
        showErrorMessageAndExit("Failed to load definitions for OpenGL functions.");
    }

    glEnable(GL_DEPTH_TEST);
    glClearColor(0.239f, 0.541f, 0.513f, 1.0f);

    // Get a texture shader program
    GLint program = CompileShader("Texture");

    // Initialize vertex buffers
    initVertexBuffers(program);

    // Initialize textures
    initTextures();
}

void draw()
{
    // Clear the screen with the background color
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Bind texture
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, g_texture);

    // Draw the cube
    glDrawElements(GL_TRIANGLES, 18, GL_UNSIGNED_SHORT, 0);

    // Swap the back and the front buffers
    glutSwapBuffers();
}

int main(int argc, char **argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
    glutInitWindowSize(g_width, g_height);
    glutInitWindowPosition(100, 100);
    glutCreateWindow("Textured Cube");

    initialize();
    glutDisplayFunc(draw);
    
    glutMainLoop();
    return 0;
}